from typing import Dict

from ..base import ToolCommandLineParser
from ..intents import ToolIntentionType

from typing import Union, Set


class PipParser(ToolCommandLineParser):
    def get_tool_name(self) -> str:
        return "pip"

    def get_command_hierarchy(self) -> Dict[str, Union[ToolIntentionType, Dict]]:
        """
        Context for command hierarchy parsing
        """
        return {
            "install": ToolIntentionType.ADD_PACKAGE,
            "uninstall": ToolIntentionType.REMOVE_PACKAGE,
            "download": ToolIntentionType.DOWNLOAD_PACKAGE,
            "wheel": ToolIntentionType.DOWNLOAD_PACKAGE,
            "query": ToolIntentionType.SEARCH_PACKAGES,
            "index": {
                "versions": ToolIntentionType.SEARCH_PACKAGES,
            },
        }

    def get_known_flags(self) -> Dict[str, Set[str]]:
        """
        Define flags that DON'T take values to avoid consuming packages
        """
        return {
            # Global flags (available for all commands)
            "global": {
                "help",
                "h",
                "debug",
                "isolated",
                "require-virtualenv",
                "verbose",
                "v",
                "version",
                "V",
                "quiet",
                "q",
                "no-input",
                "no-cache-dir",
                "disable-pip-version-check",
                "no-color",
                # Index specific
                "no-index",
            },
            # install-specific flags
            "install": {
                "no-deps",
                "pre",
                "dry-run",
                "user",
                "upgrade",
                "U",
                "force-reinstall",
                "ignore-installed",
                "I",
                "ignore-requires-python",
                "no-build-isolation",
                "use-pep517",
                "no-use-pep517",
                "check-build-dependencies",
                "break-system-packages",
                "compile",
                "no-compile",
                "no-warn-script-location",
                "no-warn-conflicts",
                "prefer-binary",
                "require-hashes",
                "no-clean",
            },
            # uninstall-specific flags
            "uninstall": {
                "yes",
                "y",
                "break-system-packages",
            },
            # download-specific flags
            "download": {
                "no-deps",
                "no-binary",
                "only-binary",
                "prefer-binary",
                "pre",
                "require-hashes",
                "no-build-isolation",
                "use-pep517",
                "no-use-pep517",
                "check-build-dependencies",
                "ignore-requires-python",
                "no-clean",
            },
            "index.versions": {
                "ignore-requires-python",
                "pre",
                "json",
                "no-index",
            },
        }



import unittest
import pytest
import json
import base64
from unittest.mock import MagicMock, patch

import typer
from safety.tool.auth import index_credentials, build_index_url
from safety.tool.constants import PYPI_PUBLIC_REPOSITORY_URL


@pytest.mark.unit
class TestIndexCredentials(unittest.TestCase):
    """
    Test cases for index_credentials function.
    """

    def test_index_credentials_with_full_auth_object(self):
        """
        Test index_credentials when ctx.obj.auth is fully populated with token and api_key.
        """
        ctx = MagicMock(spec=typer.Context)
        ctx.obj = MagicMock()
        ctx.obj.auth.client.token = {"access_token": "test_token"}
        ctx.obj.auth.client.api_key = "test_api_key"
        ctx.obj.project.id = "test_project_id"

        result = index_credentials(ctx)

        decoded = json.loads(
            base64.urlsafe_b64decode(result.encode("utf-8")).decode("utf-8")
        )

        self.assertEqual(decoded["version"], "1.0")
        self.assertEqual(decoded["access_token"], "test_token")
        self.assertEqual(decoded["api_key"], "test_api_key")
        self.assertEqual(decoded["project_id"], "test_project_id")

    def test_index_credentials_with_missing_token(self):
        """
        Test index_credentials when token is None but api_key is present.
        """
        ctx = MagicMock(spec=typer.Context)
        ctx.obj = MagicMock()
        ctx.obj.auth.client.token = None
        ctx.obj.auth.client.api_key = "test_api_key"
        ctx.obj.project.id = "test_project_id"

        result = index_credentials(ctx)

        decoded = json.loads(
            base64.urlsafe_b64decode(result.encode("utf-8")).decode("utf-8")
        )
        self.assertEqual(decoded["version"], "1.0")
        self.assertIsNone(decoded["access_token"])
        self.assertEqual(decoded["api_key"], "test_api_key")
        self.assertEqual(decoded["project_id"], "test_project_id")

    def test_index_credentials_with_missing_api_key(self):
        """
        Test index_credentials when api_key is None but token is present.
        """
        ctx = MagicMock(spec=typer.Context)
        ctx.obj = MagicMock()
        ctx.obj.auth.client.token = {"access_token": "test_token"}
        ctx.obj.auth.client.api_key = None
        ctx.obj.project.id = "test_project_id"

        result = index_credentials(ctx)

        decoded = json.loads(
            base64.urlsafe_b64decode(result.encode("utf-8")).decode("utf-8")
        )

        self.assertEqual(decoded["version"], "1.0")
        self.assertEqual(decoded["access_token"], "test_token")
        self.assertIsNone(decoded["api_key"])
        self.assertEqual(decoded["project_id"], "test_project_id")

    def test_index_credentials_with_no_auth(self):
        """
        Test index_credentials when ctx.obj.auth is None.
        """

        ctx = MagicMock(spec=typer.Context)
        ctx.obj = MagicMock()
        ctx.obj.auth = None
        ctx.obj.project.id = "test_project_id"

        result = index_credentials(ctx)

        decoded = json.loads(
            base64.urlsafe_b64decode(result.encode("utf-8")).decode("utf-8")
        )

        self.assertEqual(decoded["version"], "1.0")
        self.assertIsNone(decoded["access_token"])
        self.assertIsNone(decoded["api_key"])
        self.assertEqual(decoded["project_id"], "test_project_id")

    def test_index_credentials_with_no_project(self):
        """
        Test index_credentials when ctx.obj.project is None.
        """

        ctx = MagicMock(spec=typer.Context)
        ctx.obj = MagicMock()
        ctx.obj.auth.client.token = {"access_token": "test_token"}
        ctx.obj.auth.client.api_key = "test_api_key"
        ctx.obj.project = None

        result = index_credentials(ctx)

        decoded = json.loads(
            base64.urlsafe_b64decode(result.encode("utf-8")).decode("utf-8")
        )

        self.assertEqual(decoded["version"], "1.0")
        self.assertEqual(decoded["access_token"], "test_token")
        self.assertEqual(decoded["api_key"], "test_api_key")
        self.assertIsNone(decoded["project_id"])

    def test_index_credentials_correct_encoding(self):
        """
        Test that index_credentials correctly encodes the credentials in base64url format.
        """

        ctx = MagicMock(spec=typer.Context)
        ctx.obj = MagicMock()
        ctx.obj.auth.client.token = {"access_token": "test_token"}
        ctx.obj.auth.client.api_key = "test_api_key"
        ctx.obj.project.id = "test_project_id"

        result = index_credentials(ctx)

        expected_json = json.dumps(
            {
                "version": "1.0",
                "access_token": "test_token",
                "api_key": "test_api_key",
                "project_id": "test_project_id",
            }
        )

        expected_encoded = base64.urlsafe_b64encode(
            expected_json.encode("utf-8")
        ).decode("utf-8")

        self.assertEqual(result, expected_encoded)


@pytest.mark.unit
class TestBuildIndexUrl:
    """
    Test suite for build_index_url
    """

    def setup_method(self):
        self.ctx = MagicMock(spec=typer.Context)
        self.ctx.obj = MagicMock()

    def test_build_index_url_injects_and_defaults(self):
        """
        It injects user:<b64>@ into netloc and defaults to PYPI_PUBLIC_REPOSITORY_URL when index_url is None.
        """
        with patch(
            "safety.tool.auth.index_credentials", return_value="mock_creds"
        ) as mock_ic:
            # explicit custom URL
            custom_url = "https://pkgs.safetycli.com/repository/safety-cybersecurity/project/safety/pypi/simple/"
            result_custom = build_index_url(self.ctx, custom_url, "pypi")
            assert (
                result_custom
                == "https://user:mock_creds@pkgs.safetycli.com/repository/safety-cybersecurity/project/safety/pypi/simple/"
            )

            # default to PYPI_PUBLIC_REPOSITORY_URL
            result_default = build_index_url(self.ctx, None, "pypi")
            expected_host = PYPI_PUBLIC_REPOSITORY_URL.replace("https://", "")
            assert result_default == f"https://user:mock_creds@{expected_host}"

            # Called for both invocations
            assert mock_ic.call_count == 2

    @pytest.mark.parametrize(
        "input_url,expected",
        [
            ("https://simple.example.com/", "https://user:X@simple.example.com/"),
            (
                "http://pypi.example.com/simple/",
                "http://user:X@pypi.example.com/simple/",
            ),
            (
                "https://pypi.example.com:8080/simple/",
                "https://user:X@pypi.example.com:8080/simple/",
            ),
            (
                "https://pypi.example.com/custom/path/?param=value",
                "https://user:X@pypi.example.com/custom/path/?param=value",
            ),
            (
                "https://registry.example.com:443/v1/repositories/simple/?format=json&auth=basic",
                "https://user:X@registry.example.com:443/v1/repositories/simple/?format=json&auth=basic",
            ),
        ],
    )
    def test_build_index_url_preserves_components(self, input_url, expected):
        """
        It preserves scheme, host, port, path, and query while injecting credentials.
        """
        with patch("safety.tool.auth.index_credentials", return_value="X") as mock_ic:
            result = build_index_url(self.ctx, input_url, "pypi")
            assert result == expected
            mock_ic.assert_called_once_with(self.ctx)

    def test_build_index_url_with_existing_auth_is_prepended(self):
        """
        Current implementation prepends user:<b64>@ before existing auth if present in netloc.
        """
        with patch(
            "safety.tool.auth.index_credentials", return_value="safety_creds"
        ) as mock_ic:
            url_with_auth = "https://old_user:old_pass@pypi.example.com/simple/"
            result = build_index_url(self.ctx, url_with_auth, "pypi")
            assert (
                result
                == "https://user:safety_creds@old_user:old_pass@pypi.example.com/simple/"
            )
            mock_ic.assert_called_once_with(self.ctx)



from typing import Dict, Union, Set

from ..base import ToolCommandLineParser
from ..intents import ToolIntentionType


UV_CACHE_FLAGS = {
    "no-cache",
    "n",
    "refresh",
}

UV_PYTHON_FLAGS = {
    "managed-python",
    "no-managed-python",
    "no-python-downloads",
}


UV_INDEX_FLAGS = {
    "no-index",
}

UV_RESOLVER_FLAGS = {
    "upgrade",
    "U",
    "no-sources",
}

UV_INSTALLER_FLAGS = {
    "reinstall",
    "compile-bytecode",
}

UV_BUILD_FLAGS = {
    "no-build-isolation",
    "no-build",
    "no-binary",
}

UV_GLOBAL_FLAGS = {
    "quiet",
    "q",
    "verbose",
    "v",
    "native-tls",
    "offline",
    "no-progress",
    "no-config",
    "help",
    "h",
    "version",
    "V",
}

UV_PIP_INSTALL_FLAGS = {
    "all-extras",
    "no-deps",
    "require-hashes",
    "no-verify-hashes",
    "system",
    "break-system-packages",
    "no-break-system-packages",
    "no-build",
    "exact",
    "strict",
    "dry-run",
    "user",
}

UV_PIP_UNINSTALL_FLAGS = {
    "system",
    "break-system-packages",
    "no-break-system-packages",
    "dry-run",
}


UV_KNOWN_FLAGS: Dict[str, Set[str]] = {
    "global": UV_GLOBAL_FLAGS
    | UV_CACHE_FLAGS
    | UV_PYTHON_FLAGS
    | UV_INDEX_FLAGS
    | UV_RESOLVER_FLAGS
    | UV_INSTALLER_FLAGS
    | UV_BUILD_FLAGS,
    # 2-level commands
    "add": {
        # From `uv add --help`
        "dev",
        "editable",
        "raw",
        "no-sync",
        "locked",
        "frozen",
        "active",
        "workspace",
        "no-workspace",
        "no-install-project",
        "no-install-workspace",
    },
    "remove": {
        "dev",
        "no-sync",
        "active",
        "locked",
        "frozen",
    },
    "sync": {
        "all-extras",
        "no-dev",
        "only-dev",
        "no-default-groups",
        "all-groups",
        "no-editable",
        "inexact",
        "active",
        "no-install-project",
        "no-install-workspace",
        "locked",
        "frozen",
        "dry-run",
        "all-packages",
        "check",
    },
    # 3-level pip commands
    "pip.install": UV_PIP_INSTALL_FLAGS,
    "pip.uninstall": UV_PIP_UNINSTALL_FLAGS,
}


class UvParser(ToolCommandLineParser):
    def get_tool_name(self) -> str:
        return "uv"

    def get_command_hierarchy(self) -> Dict[str, Union[ToolIntentionType, Dict]]:
        """
        Context for command hierarchy parsing
        """
        return {
            # 2-level commands
            "add": ToolIntentionType.ADD_PACKAGE,
            "remove": ToolIntentionType.REMOVE_PACKAGE,
            "build": ToolIntentionType.BUILD_PROJECT,
            "sync": ToolIntentionType.SYNC_PACKAGES,
            # 3-level commands
            "pip": {
                "install": ToolIntentionType.ADD_PACKAGE,
                "uninstall": ToolIntentionType.REMOVE_PACKAGE,
                "download": ToolIntentionType.DOWNLOAD_PACKAGE,
                "list": ToolIntentionType.LIST_PACKAGES,
            },
        }

    def get_known_flags(self) -> Dict[str, Set[str]]:
        """
        Define flags that DON'T take values for uv.
        These were derived from `uv --help` and subcommand helps.
        """
        return UV_KNOWN_FLAGS



import logging
import shutil
import subprocess
from pathlib import Path
from typing import Optional

import typer
from rich.console import Console

from safety.tool.constants import (
    NPMJS_PUBLIC_REPOSITORY_URL,
    NPMJS_ORGANIZATION_REPOSITORY_URL,
    NPMJS_PROJECT_REPOSITORY_URL,
)
from safety.tool.resolver import get_unwrapped_command
from safety.utils.pyapp_utils import get_path, get_env

from safety.console import main_console
from safety.tool.auth import build_index_url

logger = logging.getLogger(__name__)


class Npm:
    @classmethod
    def is_installed(cls) -> bool:
        """
        Checks if the NPM program is installed

        Returns:
            True if NPM is installed on system, or false otherwise
        """
        return shutil.which("npm", path=get_path()) is not None

    @classmethod
    def configure_project(
        cls,
        project_path: Path,
        org_slug: Optional[str],
        project_id: Optional[str],
        console: Console = main_console,
    ) -> Optional[Path]:
        """
        Configures Safety index url for specified npmrc file.

        Args:
            file (Path): Path to npmrc file.
            org_slug (str): Organization slug.
            project_id (str): Project identifier.
            console (Console): Console instance.
        """
        if not cls.is_installed():
            logger.error("NPM is not installed.")
            return None

        repository_url = (
            NPMJS_PROJECT_REPOSITORY_URL.format(org_slug, project_id)
            if project_id and org_slug
            else (
                NPMJS_ORGANIZATION_REPOSITORY_URL.format(org_slug)
                if org_slug
                else NPMJS_PUBLIC_REPOSITORY_URL
            )
        )

        project_root = project_path.resolve()
        if not project_root.is_dir():
            project_root = project_path.parent

        result = subprocess.run(
            [
                get_unwrapped_command(name="npm"),
                "config",
                "set",
                "registry",
                repository_url,
                "--location",
                "project",
            ],
            capture_output=True,
            cwd=project_root,
            env=get_env(),
        )

        if result.returncode != 0:
            logger.error(
                f"Failed to configure NPM project settings: {result.stderr.decode('utf-8')}"
            )
            return None

        return project_root

    @classmethod
    def configure_system(
        cls, org_slug: Optional[str], console: Console = main_console
    ) -> Optional[Path]:
        """
        Configures NPM system to use to Safety index url.
        """
        if not cls.is_installed():
            logger.error("NPM is not installed.")
            return None

        try:
            repository_url = (
                NPMJS_ORGANIZATION_REPOSITORY_URL.format(org_slug)
                if org_slug
                else NPMJS_PUBLIC_REPOSITORY_URL
            )
            result = subprocess.run(
                [
                    get_unwrapped_command(name="npm"),
                    "config",
                    "set",
                    "-g",
                    "registry",
                    repository_url,
                ],
                capture_output=True,
                env=get_env(),
            )

            if result.returncode != 0:
                logger.error(
                    f"Failed to configure NPM global settings: {result.stderr.decode('utf-8')}"
                )
                return None

            query_config_result = subprocess.run(
                [
                    get_unwrapped_command(name="npm"),
                    "config",
                    "get",
                    "globalconfig",
                ],
                capture_output=True,
                env=get_env(),
            )
            config_file_path = query_config_result.stdout.decode("utf-8").strip()

            if config_file_path:
                return Path(config_file_path)

            logger.error("Failed to match the config file path written by NPM.")
            return Path()
        except Exception:
            logger.exception("Failed to configure NPM global settings.")

        return None

    @classmethod
    def reset_system(cls, console: Console = main_console):
        # TODO: Move this logic and implement it in a more robust way
        try:
            subprocess.run(
                [
                    get_unwrapped_command(name="npm"),
                    "config",
                    "set",
                    "-g",
                    "registry",
                ],
                capture_output=True,
                env=get_env(),
            )
        except Exception:
            console.print("Failed to reset NPM global settings.")

    @classmethod
    def build_index_url(cls, ctx: typer.Context, index_url: Optional[str]) -> str:
        return build_index_url(ctx, index_url, "npm")
